"""Basic mathematical functionalities of the package.

Defines the basic mathematical operations for classes defined in this
package. FDataBasis and FDataGrid.

"""
import numpy
import scipy.integrate
from ..exploratory.stats import mean


__author__ = "Miguel Carbajo Berrocal"
__license__ = "GPL3"
__version__ = ""
__maintainer__ = ""
__email__ = ""
__status__ = "Development"


def sqrt(fdatagrid):
    """Perform a element wise square root operation.

    Args:
        fdatagrid (FDataGrid): Object to whose elements the square root
            operation is going to be applied.

    Returns:
        FDataGrid: Object whose elements are the square roots of the original.

    """
    return fdatagrid.copy(data_matrix=numpy.sqrt(fdatagrid.data_matrix))


def absolute(fdatagrid):
    """Get the absolute value of all elements in the FDataGrid object.

    Args:
        fdatagrid (FDataGrid): Object from whose elements the absolute value
            is going to be retrieved.

    Returns:
        FDataGrid: Object whose elements are the absolute values of the
            original.

    """
    return fdatagrid.copy(data_matrix=numpy.absolute(fdatagrid.data_matrix))


def round(fdatagrid, decimals=0):
    """Round all elements of the object.

    Args:
        fdatagrid (FDataGrid): Object to whose elements are going to be
            rounded.
        decimals (int, optional): Number of decimals wanted. Defaults to 0.

    Returns:
        FDataGrid: Object whose elements are rounded.

    """
    return fdatagrid.round(decimals)


def exp(fdatagrid):
    """Perform a element wise exponential operation.

    Args:
        fdatagrid (FDataGrid): Object to whose elements the exponential
            operation is going to be applied.

    Returns:
        FDataGrid: Object whose elements are the result of exponentiating
            the elements of the original.

    """
    return fdatagrid.copy(data_matrix=numpy.exp(fdatagrid.data_matrix))


def log(fdatagrid):
    """Perform a element wise logarithm operation.

    Args:
        fdatagrid (FDataGrid): Object to whose elements the logarithm
            operation is going to be applied.

    Returns:
        FDataGrid: Object whose elements are the logarithm of the original.

    """
    return fdatagrid.copy(data_matrix=numpy.log(fdatagrid.data_matrix))


def log10(fdatagrid):
    """Perform an element wise base 10 logarithm operation.

    Args:
        fdatagrid (FDataGrid): Object to whose elements the base 10 logarithm
            operation is going to be applied.

    Returns:
        FDataGrid: Object whose elements are the base 10 logarithm of the
            original.

    """
    return fdatagrid.copy(data_matrix=numpy.log10(fdatagrid.data_matrix))


def log2(fdatagrid):
    """Perform an element wise binary logarithm operation.

    Args:
        fdatagrid (FDataGrid): Object to whose elements the binary logarithm
            operation is going to be applied.

    Returns:
        FDataGrid: Object whose elements are the binary logarithm of the
            original.

    """
    return fdatagrid.copy(data_matrix=numpy.log2(fdatagrid.data_matrix))


def cumsum(fdatagrid):
    """Return the cumulative sum of the samples.

    Args:
        fdatagrid (FDataGrid): Object over whose samples the cumulative sum is
            going to be calculated.

    Returns:
        FDataGrid: Object with the sample wise cumulative sum.

    """
    return fdatagrid.copy(data_matrix=numpy.cumsum(fdatagrid.data_matrix,
                                                   axis=0))


def inner_product(fdatagrid, fdatagrid2):
    r"""Return inner product for FDataGrid.

    Calculates the inner product amongst all the samples in two
    FDataGrid objects.

    For each pair of samples f and g the inner product is defined as:

    .. math::
        <f, g> = \int_a^bf(x)g(x)dx

    The integral is approximated using Simpson's rule.

    Args:
        fdatagrid (FDataGrid): First FDataGrid object.
        fdatagrid2 (FDataGrid): Second FDataGrid object.

    Returns:
        numpy.darray: Matrix with as many rows as samples in the first
        object and as many columns as samples in the second one. Each
        element (i, j) of the matrix is the inner product of the ith sample
        of the first object and the jth sample of the second one.

    Examples:
        The inner product of the :math:'f(x) = x` and the constant
        :math:`y=1` defined over the interval [0,1] is the area of the
        triangle delimited by the the lines y = 0, x = 1 and y = x; 0.5.

        >>> import skfda
        >>> x = numpy.linspace(0,1,1001)
        >>> fd1 = skfda.FDataGrid(x,x)
        >>> fd2 = skfda.FDataGrid(numpy.ones(len(x)),x)
        >>> inner_product(fd1, fd2)
        array([[ 0.5]])

        If the FDataGrid object contains more than one sample

        >>> fd1 = skfda.FDataGrid([x, numpy.ones(len(x))], x)
        >>> fd2 = skfda.FDataGrid([numpy.ones(len(x)), x] ,x)
        >>> inner_product(fd1, fd2).round(2)
        array([[ 0.5 , 0.33],
               [ 1.  , 0.5 ]])

    """
    if fdatagrid.ndim_domain != 1:
        raise NotImplementedError("This method only works when the dimension "
                                  "of the domain of the FDatagrid object is "
                                  "one.")
    # Checks
    if not numpy.array_equal(fdatagrid.sample_points,
                             fdatagrid2.sample_points):
        raise ValueError("Sample points for both objects must be equal")

    # Creates an empty matrix with the desired size to store the results.
    matrix = numpy.empty([fdatagrid.nsamples, fdatagrid2.nsamples])
    # Iterates over the different samples of both objects.
    for i in range(fdatagrid.nsamples):
        for j in range(fdatagrid2.nsamples):
            # Calculates the inner product using Simpson's rule.
            matrix[i, j] = (scipy.integrate.simps(
                fdatagrid.data_matrix[i, ..., 0] *
                fdatagrid2.data_matrix[j, ..., 0],
                x=fdatagrid.sample_points[0]
                ))
    return matrix



"""Defines methods for the validation of the smoothing."""
import numpy

from . import kernel_smoothers


__author__ = "Miguel Carbajo Berrocal"
__email__ = "miguel.carbajo@estudiante.uam.es"


def cv(fdatagrid, s_matrix):
    r"""Cross validation scoring method.

    It calculates the cross validation score for every sample in a FDataGrid
    object given a smoothing matrix :math:`\hat{H}^\nu` calculated with a
    parameter :math:`\nu`:

    .. math::
        CV(\nu)=\frac{1}{n} \sum_i \left(y_i - \hat{y}_i^{\nu(
        -i)}\right)^2

    Where :math:`\hat{y}_i^{\nu(-i)}` is the adjusted :math:`y_i` when the
    the pair of values :math:`(x_i,y_i)` are excluded in the smoothing. This
    would require to recalculate the smoothing matrix n times. Fortunately
    the above formula can be expressed in a way where the smoothing matrix
    does not need to be calculated again.

    .. math::
        CV(\nu)=\frac{1}{n} \sum_i \left(\frac{y_i - \hat{y}_i^\nu}{1 -
        \hat{H}_{ii}^\nu}\right)^2

    Args:
        fdatagrid (FDataGrid): Object over which the CV score is calculated.
        s_matrix (numpy.darray): Smoothig matrix.

    Returns:
        float: Cross validation score.

    """
    y = fdatagrid.data_matrix[..., 0]
    y_est = numpy.dot(s_matrix, y.T).T
    return numpy.mean(((y - y_est) / (1 - s_matrix.diagonal())) ** 2)


def gcv(fdatagrid, s_matrix, penalisation_function=None):
    r"""General cross validation scoring method.

    It calculates the general cross validation score for every sample in a
    FDataGrid object given a smoothing matrix :math:`\hat{H}^\nu`
    calculated with a parameter :math:`\nu`:

    .. math::
        GCV(\nu)=\Xi(\nu,n)\frac{1}{n} \sum_i \left(y_i - \hat{
        y}_i^\nu\right)^2

    Where :math:`\hat{y}_i^{\nu}` is the adjusted :math:`y_i` and
    :math:`\Xi` is a penalisation function. By default the penalisation
    function is:

    .. math::
        \Xi(\nu,n) = \left( 1 - \frac{tr(\hat{H}^\nu)}{n} \right)^{-2}

    But others such as the Akaike's information criterion can be considered.

    Args:
        fdatagrid (FDataGrid): Object over which the CV score is calculated.
        s_matrix (numpy.darray): Smoothig matrix.
        penalisation_function (Function): Function taking a smoothing matrix
            and returing a penalisation score. If None the general cross
            validation penalisation is applied. Defaults to None.

    Returns:
        float: Cross validation score.

    """
    y = fdatagrid.data_matrix[..., 0]
    y_est = numpy.dot(s_matrix, y.T).T
    if penalisation_function is not None:
        return (numpy.mean(((y - y_est) / (1 - s_matrix.diagonal())) ** 2)
                * penalisation_function(s_matrix))
    return (numpy.mean(((y - y_est) / (1 - s_matrix.diagonal())) ** 2)
            * (1 - s_matrix.diagonal().mean()) ** -2)


def minimise(fdatagrid, parameters,
             smoothing_method=kernel_smoothers.nw, cv_method=gcv,
             penalisation_function=None, **kwargs):
    """Chooses the best smoothness parameter and performs smoothing.

    Performs the smoothing of a FDataGrid object choosing the best
    parameter of a given list using a cross validation scoring method.

    Args:
        fdatagrid (FDataGrid): FDataGrid object.
        parameters (list of double): List of parameters to be tested.
        smoothing_method (Function): Function that takes a list of
            discretised points, a parameter, an optionally a weights matrix
            and returns a hat matrix or smoothing matrix.
        cv_method (Function): Function that takes a matrix,
            a smoothing matrix, and optionally a weights matrix and
            calculates a cross validation score.
        penalisation_function(Fuction): if gcv is selected as cv_method a
            penalisation function can be specified through this parameter.

    Returns:
        dict: A dictionary containing the following:

            {
                'scores': (list of double) List of the scores for each
                    parameter.

                'best_score': (double) Minimum score.

                'best_parameter': (double) Parameter that produces the
                    lesser score.

                'hat_matrix': (numpy.darray) Hat matrix built with the best
                    parameter.

                'fdatagrid': (FDataGrid) Smoothed FDataGrid object.

            }

    Examples:
        Creates a FDataGrid object of the function :math:`y=x^2` and peforms
        smoothing by means of the k-nearest neighbours method.

        >>> import skfda
        >>> x = numpy.linspace(-2, 2, 5)
        >>> fd = skfda.FDataGrid(x ** 2, x)
        >>> res = minimise(fd, [2,3], smoothing_method=kernel_smoothers.knn)
        >>> numpy.array(res['scores']).round(2)
        array([ 11.67, 12.37])
        >>> round(res['best_score'], 2)
        11.67
        >>> res['best_parameter']
        2
        >>> res['hat_matrix'].round(2)
        array([[ 0.5 , 0.5 , 0.  , 0.  , 0.  ],
               [ 0.33, 0.33, 0.33, 0.  , 0.  ],
               [ 0.  , 0.33, 0.33, 0.33, 0.  ],
               [ 0.  , 0.  , 0.33, 0.33, 0.33],
               [ 0.  , 0.  , 0.  , 0.5 , 0.5 ]])
        >>> res['fdatagrid'].round(2)
        FDataGrid(
            array([[[ 2.5 ],
                    [ 1.67],
                    [ 0.67],
                    [ 1.67],
                    [ 2.5 ]]]),
            sample_points=[array([-2., -1.,  0.,  1.,  2.])],
            domain_range=array([[-2.,  2.]]),
            ...)

        Other validation methods can be used such as cross-validation or
        general cross validation using other penalisation functions.

        >>> res = minimise(fd, [2,3], smoothing_method=kernel_smoothers.knn,
        ...                cv_method=cv)
        >>> numpy.array(res['scores']).round(2)
        array([ 4.2, 5.5])
        >>> res = minimise(fd, [2,3], smoothing_method=kernel_smoothers.knn,
        ...                penalisation_function=aic)
        >>> numpy.array(res['scores']).round(2)
        array([ 9.35, 10.71])
        >>> res = minimise(fd, [2,3], smoothing_method=kernel_smoothers.knn,
        ...                penalisation_function=fpe)
        >>> numpy.array(res['scores']).round(2)
        array([ 9.8, 11. ])
        >>> res = minimise(fd, [2,3], smoothing_method=kernel_smoothers.knn,
        ...                penalisation_function=shibata)
        >>> numpy.array(res['scores']).round(2)
        array([ 7.56, 9.17])
        >>> res = minimise(fd, [2,3], smoothing_method=kernel_smoothers.knn,
        ...                penalisation_function=rice)
        >>> numpy.array(res['scores']).round(2)
        array([ 21. , 16.5])

    """
    if fdatagrid.ndim_domain != 1:
        raise NotImplementedError("This method only works when the dimension "
                                  "of the domain of the FDatagrid object is "
                                  "one.")
    if fdatagrid.ndim_image != 1:
        raise NotImplementedError("This method only works when the dimension "
                                  "of the image of the FDatagrid object is "
                                  "one.")
    # Reduce one dimension the sample points.
    sample_points = fdatagrid.sample_points[0]
    scores = []
    # Calculates the scores for each parameter.
    if penalisation_function is not None:
        for h in parameters:
            s = smoothing_method(sample_points, h, **kwargs)
            scores.append(
                cv_method(fdatagrid, s,
                          penalisation_function=penalisation_function))
    else:
        for h in parameters:
            s = smoothing_method(sample_points, h, **kwargs)
            scores.append(
                cv_method(fdatagrid, s))
    # gets the best parameter.
    h = parameters[int(numpy.argmin(scores))]
    s = smoothing_method(sample_points, h, **kwargs)
    fdatagrid_adjusted = fdatagrid.copy(
        data_matrix=numpy.dot(fdatagrid.data_matrix[..., 0], s.T))

    return {'scores': scores,
            'best_score': numpy.min(scores),
            'best_parameter': h,
            'hat_matrix': s,
            'fdatagrid': fdatagrid_adjusted,
            }


def aic(s_matrix):
    r"""Akaike's information criterion for cross validation.

    .. math::
        \Xi(\nu,n) = \exp\left(2 * \frac{tr(\hat{H}^\nu)}{n}\right)

    Args:
        s_matrix (numpy.darray): Smoothing matrix whose penalisation
            score is desired.

    Returns:
         float: Penalisation given by the Akaike's information criterion.

    """
    return numpy.exp(2 * s_matrix.diagonal().mean())


def fpe(s_matrix):
    r"""Finite prediction error for cross validation.

    .. math::
        \Xi(\nu,n) = \frac{1 + \frac{tr(\hat{H}^\nu)}{n}}{1 -
        \frac{tr(\hat{H}^\nu)}{n}}

    Args:
        s_matrix (numpy.darray): Smoothing matrix whose penalisation
            score is desired.

    Returns:
         float: Penalisation given by the finite prediction error.

    """
    return (1 + s_matrix.diagonal().mean()) / (1 - s_matrix.diagonal().mean())


def shibata(s_matrix):
    r"""Shibata's model selector for cross validation.

    .. math::
        \Xi(\nu,n) = 1 + 2 * \frac{tr(\hat{H}^\nu)}{n}

    Args:
        s_matrix (numpy.darray): Smoothing matrix whose penalisation
            score is desired.

    Returns:
         float: Penalisation given by the Shibata's model selector.

    """
    return 1 + 2 * s_matrix.diagonal().mean()


def rice(s_matrix):
    r"""Rice's bandwidth selector for cross validation.

    .. math::
        \Xi(\nu,n) = \left(1 - 2 * \frac{tr(\hat{H}^\nu)}{n}\right)^{-1}

    Args:
        s_matrix (numpy.darray): Smoothing matrix whose penalisation
            score is desired.

    Returns:
         float: Penalisation given by the Rice's bandwidth selector.

    """
    return (1 - 2 * s_matrix.diagonal().mean()) ** -1



"""Test to check the extrapolation module"""

import unittest

import numpy as np

from skfda.datasets import make_sinusoidal_process
from skfda.representation.basis import Fourier
from skfda import FDataGrid, FDataBasis
from skfda.representation.extrapolation import (
    PeriodicExtrapolation, BoundaryExtrapolation, ExceptionExtrapolation,
    FillExtrapolation)


class TestBasis(unittest.TestCase):

    def setUp(self):
        self.grid = make_sinusoidal_process(n_samples=2, random_state=0)
        self.basis = self.grid.to_basis(Fourier())
        self.dummy_data = [[1, 2, 3], [2, 3, 4]]

    def test_constructor_FDataBasis_setting(self):
        coeff = self.dummy_data
        basis = Fourier(nbasis=3)

        a = FDataBasis(basis, coeff)
        np.testing.assert_equal(a.extrapolation, None)

        a = FDataBasis(basis, coeff, extrapolation=PeriodicExtrapolation())
        np.testing.assert_equal(a.extrapolation, PeriodicExtrapolation())

        np.testing.assert_equal(
            a.extrapolation == BoundaryExtrapolation(), False)

        a = FDataBasis(basis, coeff, extrapolation=BoundaryExtrapolation())
        np.testing.assert_equal(a.extrapolation, BoundaryExtrapolation())

        a = FDataBasis(basis, coeff, extrapolation=ExceptionExtrapolation())
        np.testing.assert_equal(a.extrapolation, ExceptionExtrapolation())

        a = FDataBasis(basis, coeff, extrapolation=FillExtrapolation(0))
        np.testing.assert_equal(a.extrapolation, FillExtrapolation(0))

        np.testing.assert_equal(a.extrapolation == FillExtrapolation(1), False)

    def test_FDataBasis_setting(self):
        coeff = self.dummy_data
        basis = Fourier(nbasis=3)
        a = FDataBasis(basis, coeff)

        a.extrapolation = "periodic"
        np.testing.assert_equal(a.extrapolation, PeriodicExtrapolation())

        a.extrapolation = "bounds"
        np.testing.assert_equal(a.extrapolation, BoundaryExtrapolation())

        a.extrapolation = "exception"
        np.testing.assert_equal(a.extrapolation, ExceptionExtrapolation())

        a.extrapolation = "zeros"
        np.testing.assert_equal(a.extrapolation, FillExtrapolation(0))

        a.extrapolation = "nan"
        np.testing.assert_equal(a.extrapolation, FillExtrapolation(np.nan))

    def test_constructor_FDataGrid_setting(self):
        data = self.dummy_data

        a = FDataGrid(data)
        np.testing.assert_equal(a.extrapolation, None)

        a = FDataGrid(data, extrapolation=PeriodicExtrapolation())
        np.testing.assert_equal(a.extrapolation, PeriodicExtrapolation())

        a = FDataGrid(data, extrapolation=BoundaryExtrapolation())
        np.testing.assert_equal(a.extrapolation, BoundaryExtrapolation())

        np.testing.assert_equal(
            a.extrapolation == ExceptionExtrapolation(), False)

        a = FDataGrid(data, extrapolation=ExceptionExtrapolation())
        np.testing.assert_equal(a.extrapolation, ExceptionExtrapolation())

        a = FDataGrid(data, extrapolation=FillExtrapolation(0))
        np.testing.assert_equal(a.extrapolation, FillExtrapolation(0))
        np.testing.assert_equal(a.extrapolation == FillExtrapolation(1), False)

    def test_FDataGrid_setting(self):
        data = self.dummy_data
        a = FDataGrid(data)

        a.extrapolation = PeriodicExtrapolation()
        np.testing.assert_equal(a.extrapolation, PeriodicExtrapolation())

        a.extrapolation = "bounds"
        np.testing.assert_equal(a.extrapolation, BoundaryExtrapolation())

        a.extrapolation = "exception"
        np.testing.assert_equal(a.extrapolation, ExceptionExtrapolation())

        a.extrapolation = "zeros"
        np.testing.assert_equal(a.extrapolation, FillExtrapolation(0))

        np.testing.assert_equal(a.extrapolation == FillExtrapolation(1), False)

    def test_periodic(self):
        self.grid.extrapolation = PeriodicExtrapolation()
        data = self.grid([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[-0.724,  0.976, -0.724],
                                              [-1.086,  0.759, -1.086]])

        self.basis.extrapolation = "periodic"
        data = self.basis([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[-0.69,  0.692, -0.69],
                                              [-1.021,  1.056, -1.021]])

    def test_boundary(self):
        self.grid.extrapolation = "bounds"
        data = self.grid([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[0.976,  0.976,  0.797],
                                              [0.759,  0.759,  1.125]])

        self.basis.extrapolation = "bounds"
        data = self.basis([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[0.692, 0.692, 0.692],
                                              [1.056, 1.056, 1.056]])

    def test_exception(self):
        self.grid.extrapolation = "exception"

        with np.testing.assert_raises(ValueError):
            self.grid([-.5, 0, 1.5])

        self.basis.extrapolation = "exception"

        with np.testing.assert_raises(ValueError):
            self.basis([-.5, 0, 1.5])

    def test_zeros(self):
        self.grid.extrapolation = "zeros"
        data = self.grid([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[0.,  0.976,  0.],
                                              [0.,  0.759,  0.]])

        self.basis.extrapolation = "zeros"
        data = self.basis([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[0, 0.692, 0],
                                              [0, 1.056, 0]])

    def test_nan(self):
        self.grid.extrapolation = "nan"
        data = self.grid([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[np.nan,  0.976,  np.nan],
                                              [np.nan,  0.759,  np.nan]])

        self.basis.extrapolation = "nan"
        data = self.basis([-.5, 0, 1.5]).round(3)

        np.testing.assert_almost_equal(data, [[np.nan, 0.692, np.nan],
                                              [np.nan, 1.056, np.nan]])


if __name__ == '__main__':
    print()
    unittest.main()




import unittest
from skfda import FDataGrid
from skfda.representation.interpolation import SplineInterpolator
import numpy as np

# TODO: Unitest for grids with domain dimension > 1

class TestEvaluationSpline1_1(unittest.TestCase):
    """Test the evaluation of a grid spline interpolator with
    domain and image dimension equal to 1.
    """

    def setUp(self):
        # Data matrix of a datagrid with a dimension of domain and image equal
        # to 1.

        # Matrix of functions (x**2, (9-x)**2)
        self.data_matrix_1_1 = [np.arange(10)**2,
                                np.arange(start=9, stop=-1, step=-1)**2]


    def test_evaluation_linear_simple(self):
        """Test basic usage of evaluation"""

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10))

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(
            f(np.arange(10)), self.data_matrix_1_1)

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f([0.5,1.5,2.5]),
                                      np.array([[ 0.5,  2.5,  6.5],
                                                [72.5, 56.5, 42.5]]))

    def test_evaluation_linear_point(self):
        """Test the evaluation of a single point"""

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10))

        # Test a single point
        np.testing.assert_array_almost_equal(f(5.3).round(1),
                                      np.array([[28.3], [13.9]]))
        np.testing.assert_array_almost_equal(f([3]), np.array([[9.], [36.]]))
        np.testing.assert_array_almost_equal(f((2,)), np.array([[4.], [49.]]))


    def test_evaluation_linear_derivative(self):
        """Test derivative"""
        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10))

        # Derivate = [2*x, 2*(9-x)]
        np.testing.assert_array_almost_equal(
            f([0.5,1.5,2.5], derivative=1).round(3),
                                      np.array([[  1.,   3.,   5.],
                                                [-17., -15., -13.]]))

    def test_evaluation_linear_grid(self):
        """Test grid evaluation. With domain dimension = 1"""

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10))

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(np.arange(10)),
                                             self.data_matrix_1_1)

        res = np.array([[ 0.5,  2.5,  6.5], [72.5, 56.5, 42.5]])
        t = [0.5,1.5,2.5]

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f(t, grid=True), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res)
        np.testing.assert_array_almost_equal(f([t], grid=True), res)
        # Single point with grid
        np.testing.assert_array_almost_equal(f(3, grid=True),
                                             np.array([[9.], [36.]]))

        # Check erroneous axis
        with np.testing.assert_raises(ValueError): f((t,t), grid=True)

    def test_evaluation_linear_composed(self):

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10))

        # Evaluate (x**2, (9-x)**2) in (1,8)
        np.testing.assert_array_almost_equal(f([[1],[8]],
                                               aligned_evaluation=False),
                                      np.array([[1.], [1.]]))

        t = np.linspace(4,6,4)
        np.testing.assert_array_almost_equal(
            f([t,9-t], aligned_evaluation=False).round(2),
            np.array([[16.  , 22.  , 28.67, 36.  ],
                      [16.  , 22.  , 28.67, 36.  ]]))

        # Same length than nsample
        t = np.linspace(4,6,2)
        np.testing.assert_array_almost_equal(
            f([t,9-t], aligned_evaluation=False).round(2),
            np.array([[16.  , 36.], [16.  , 36.]]))

    def test_evaluation_linear_keepdims(self):
        """Test parameter keepdims"""

        # Default keepdims = False
        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      keepdims=False)

        # Default keepdims = True
        fk = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                       keepdims=True)

        t = [0.5,1.5,2.5]
        res = np.array([[ 0.5,  2.5,  6.5], [72.5, 56.5, 42.5]])
        res_keepdims = res.reshape((2,3,1))


        # Test combinations of keepdims with list
        np.testing.assert_array_almost_equal(f(t), res)
        np.testing.assert_array_almost_equal(f(t, keepdims=False), res)
        np.testing.assert_array_almost_equal(f(t, keepdims=True), res_keepdims)

        np.testing.assert_array_almost_equal(fk(t), res_keepdims)
        np.testing.assert_array_almost_equal(fk(t, keepdims=False), res)
        np.testing.assert_array_almost_equal(fk(t, keepdims=True), res_keepdims)

        t2 = 4
        res2 = np.array([[16.], [25.]])
        res2_keepdims = res2.reshape(2,1,1)

        # Test combinations of keepdims with a single point
        np.testing.assert_array_almost_equal(f(t2), res2)
        np.testing.assert_array_almost_equal(f(t2, keepdims=False), res2)
        np.testing.assert_array_almost_equal(f(t2, keepdims=True), res2_keepdims)

        np.testing.assert_array_almost_equal(fk(t2), res2_keepdims)
        np.testing.assert_array_almost_equal(fk(t2, keepdims=False), res2)
        np.testing.assert_array_almost_equal(fk(t2, keepdims=True), res2_keepdims)

    def test_evaluation_composed_linear_keepdims(self):
        """Test parameter keepdims with composed evaluation"""

        # Default keepdims = False
        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      keepdims=False)

        # Default keepdims = True
        fk = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                       keepdims=True)

        t = np.array([1, 2, 3])
        t = [t, 9 -  t]
        res = np.array([[ 1.,  4.,  9.], [ 1.,  4.,  9.]])
        res_keepdims = res.reshape((2,3,1))

        # Test combinations of keepdims with list
        np.testing.assert_array_almost_equal(f(t, aligned_evaluation=False), res)
        np.testing.assert_array_almost_equal(f(t, aligned_evaluation=False,
                                        keepdims=False), res)
        np.testing.assert_array_almost_equal(f(t, aligned_evaluation=False,
                                        keepdims=True), res_keepdims)

        np.testing.assert_array_almost_equal(fk(t, aligned_evaluation=False),
                                      res_keepdims)
        np.testing.assert_array_almost_equal(fk(t, aligned_evaluation=False,
                                         keepdims=False), res)
        np.testing.assert_array_almost_equal(fk(t, aligned_evaluation=False,
                                         keepdims=True), res_keepdims)

    def test_evaluation_grid_linear_keepdims(self):
        """Test grid evaluation with keepdims"""

        # Default keepdims = False
        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      keepdims=False)

        # Default keepdims = True
        fk = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                       keepdims=True)

        t = [0.5,1.5,2.5]
        res = np.array([[ 0.5,  2.5,  6.5], [72.5, 56.5, 42.5]])
        res_keepdims = res.reshape(2,3,1)

        np.testing.assert_array_almost_equal(f(t, grid=True), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True, keepdims=True),
                                      res_keepdims)
        np.testing.assert_array_almost_equal(f([t], grid=True, keepdims=False), res)

        np.testing.assert_array_almost_equal(fk(t, grid=True), res_keepdims)
        np.testing.assert_array_almost_equal(fk((t,), grid=True, keepdims=True),
                                      res_keepdims)
        np.testing.assert_array_almost_equal(fk([t], grid=True, keepdims=False), res)

    def test_evaluation_cubic_simple(self):
        """Test basic usage of evaluation"""

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      interpolator=SplineInterpolator(3))

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(np.arange(10)).round(1),
                                      self.data_matrix_1_1)

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f([0.5,1.5,2.5]).round(2),
                                      np.array([[ 0.25,  2.25,  6.25],
                                                [72.25, 56.25, 42.25]]))

    def test_evaluation_cubic_point(self):
        """Test the evaluation of a single point"""

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      interpolator=SplineInterpolator(3))

        # Test a single point
        np.testing.assert_array_almost_equal(f(5.3).round(3), np.array([[28.09],
                                                                 [13.69]]))

        np.testing.assert_array_almost_equal(f([3]).round(3), np.array([[9.], [36.]]))
        np.testing.assert_array_almost_equal(f((2,)).round(3), np.array([[4.], [49.]]))


    def test_evaluation_cubic_derivative(self):
        """Test derivative"""
        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      interpolator=SplineInterpolator(3))

        # Derivate = [2*x, 2*(9-x)]
        np.testing.assert_array_almost_equal(f([0.5,1.5,2.5], derivative=1).round(3),
                                      np.array([[  1.,   3.,   5.],
                                                [-17., -15., -13.]]))

    def test_evaluation_cubic_grid(self):
        """Test grid evaluation. With domain dimension = 1"""

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      interpolator=SplineInterpolator(3))

        t = [0.5,1.5,2.5]
        res = np.array([[ 0.25,  2.25,  6.25], [72.25, 56.25, 42.25]])


        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f(t, grid=True).round(3), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True).round(3), res)
        np.testing.assert_array_almost_equal(f([t], grid=True).round(3), res)
        # Single point with grid
        np.testing.assert_array_almost_equal(f(3, grid=True), np.array([[9.], [36.]]))

        # Check erroneous axis
        with np.testing.assert_raises(ValueError): f((t,t), grid=True)

    def test_evaluation_cubic_composed(self):

        f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                      interpolator=SplineInterpolator(3))


        # Evaluate (x**2, (9-x)**2) in (1,8)
        np.testing.assert_array_almost_equal(f([[1],[8]], aligned_evaluation=False).round(3)
                                      ,np.array([[1.], [1.]]))

        t = np.linspace(4,6,4)
        np.testing.assert_array_almost_equal(f([t,9-t], aligned_evaluation=False).round(2),
                                      np.array([[16.  , 21.78, 28.44, 36.  ],
                                                [16.  , 21.78, 28.44, 36.  ]]))

        # Same length than nsample
        t = np.linspace(4,6,2)
        np.testing.assert_array_almost_equal(f([t,9-t], aligned_evaluation=False).round(3),
                                      np.array([[16.  , 36.], [16.  , 36.]]))

    def test_evaluation_nodes(self):
        """Test interpolation in nodes for all dimensions"""

        for degree in range(1,6):
            interpolator = SplineInterpolator(degree)

            f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                          interpolator=interpolator)

            # Test interpolation in nodes
            np.testing.assert_array_almost_equal(f(np.arange(10)).round(5),
                                          self.data_matrix_1_1)

    def test_error_degree(self):


        with np.testing.assert_raises(ValueError):
            interpolator = SplineInterpolator(7)
            f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                          interpolator=interpolator)
            f(1)

        with np.testing.assert_raises(ValueError):
            interpolator = SplineInterpolator(0)
            f = FDataGrid(self.data_matrix_1_1, sample_points=np.arange(10),
                          interpolator=interpolator)
            f(1)


class TestEvaluationSpline1_n(unittest.TestCase):
    """Test the evaluation of a grid spline interpolator with
    domain dimension equal to 1 and arbitary image dimension.
    """

    def setUp(self):
        # Data matrix of a datagrid with a dimension of domain and image equal
        # to 1.

        # Matrix of functions (x**2, (9-x)**2)

        self.t = np.arange(10)

        data_1 = np.array([np.arange(10)**2,
                           np.arange(start=9, stop=-1, step=-1)**2])
        data_2 = np.sin(np.pi/81 * data_1)

        self.data_matrix_1_n = np.dstack((data_1,data_2))

        self.interpolator = SplineInterpolator(interpolation_order=2)


    def test_evaluation_simple(self):
        """Test basic usage of evaluation"""

        f = FDataGrid(self.data_matrix_1_n, sample_points=np.arange(10),
                      interpolator=self.interpolator)

        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(self.t), self.data_matrix_1_n)
        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f([1.5,2.5,3.5]),
                                             np.array([[[ 2.25    ,  0.087212],
                                                        [ 6.25    ,  0.240202],
                                                        [12.25    ,  0.45773 ]],
                                                       [[56.25    ,  0.816142],
                                                        [42.25    ,  0.997589],
                                                        [30.25    ,  0.922146]]]
                                                      )
                                             )

    def test_evaluation_point(self):
        """Test the evaluation of a single point"""

        f = FDataGrid(self.data_matrix_1_n, sample_points=np.arange(10),
                      interpolator=self.interpolator)

        # Test a single point
        np.testing.assert_array_almost_equal(f(5.3),
                                             np.array([[[28.09    ,  0.885526]],
                                                       [[13.69    ,  0.50697 ]]]
                                                      )
                                             )

    def test_evaluation_derivative(self):
        """Test derivative"""
        f = FDataGrid(self.data_matrix_1_n, sample_points=self.t,
                      interpolator=self.interpolator)

        # [(2*x, d/dx sin(pi/81*x**2)), (2*(9-x), d/dx sin(pi/81*(9-x)**2))]
        np.testing.assert_array_almost_equal(f([1.5,2.5,3.5], derivative=1),
                                             np.array([[[  3.   , 0.1162381],
                                                        [  5.   , 0.1897434],
                                                        [  7.   , 0.2453124]],
                                                       [[-15.   , 0.3385772],
                                                        [-13.   , 0.0243172],
                                                        [-11.   ,-0.1752035]]]))

    def test_evaluation_grid(self):
        """Test grid evaluation. With domain dimension = 1"""

        f = FDataGrid(self.data_matrix_1_n, sample_points=np.arange(10),
                      interpolator=SplineInterpolator(2))

        t = [1.5,2.5,3.5]
        res = np.array([[[ 2.25      ,  0.08721158],
                         [ 6.25      ,  0.24020233],
                         [12.25      ,  0.4577302 ]],
                        [[56.25      ,  0.81614206],
                         [42.25      ,  0.99758925],
                         [30.25      ,  0.92214607]]])

        # Test evaluation in a list of times
        np.testing.assert_array_almost_equal(f(t, grid=True), res)
        np.testing.assert_array_almost_equal(f((t,), grid=True), res)
        np.testing.assert_array_almost_equal(f([t], grid=True), res)

        # Check erroneous axis
        with np.testing.assert_raises(ValueError): f((t,t), grid=True)

    def test_evaluation_composed(self):

        f = FDataGrid(self.data_matrix_1_n, sample_points=self.t,
                      interpolator=self.interpolator)


        # Evaluate (x**2, (9-x)**2) in (1,8)
        np.testing.assert_array_almost_equal(f([[1],[4]],
                                               aligned_evaluation=False)[0],
                                             f(1)[0])
        np.testing.assert_array_almost_equal(f([[1],[4]],
                                               aligned_evaluation=False)[1],
                                             f(4)[1])


    def test_evaluation_keepdims(self):
        """Test keepdims"""

        f = FDataGrid(self.data_matrix_1_n, sample_points=np.arange(10),
                      interpolator=self.interpolator, keepdims=True)

        fk = FDataGrid(self.data_matrix_1_n, sample_points=np.arange(10),
                      interpolator=self.interpolator, keepdims=False)

        res = f(self.t)
        # Test interpolation in nodes
        np.testing.assert_array_almost_equal(f(self.t, keepdims=False), res)
        np.testing.assert_array_almost_equal(f(self.t, keepdims=True), res)
        np.testing.assert_array_almost_equal(fk(self.t), res)
        np.testing.assert_array_almost_equal(fk(self.t, keepdims=False), res)
        np.testing.assert_array_almost_equal(fk(self.t, keepdims=True), res)


    def test_evaluation_nodes(self):
        """Test interpolation in nodes for all dimensions"""

        for degree in range(1,6):
            interpolator = SplineInterpolator(degree)

            f = FDataGrid(self.data_matrix_1_n, sample_points=np.arange(10),
                          interpolator=interpolator)

            # Test interpolation in nodes
            np.testing.assert_array_almost_equal(f(np.arange(10)),
                                                 self.data_matrix_1_n)






if __name__ == '__main__':
    print()
    unittest.main()



"""Functional data descriptive statistics.
"""


def mean(fdata):
    """Compute the mean of all the samples in a FData object.

    Computes the mean of all the samples in a FDataGrid or FDataBasis object.

    Args:
        fdata(FDataGrid or FDataBasis): Object containing all the samples
        whose mean
            is wanted.

    Returns:
        FDataGrid or FDataBasis: A FDataGrid or FDataBasis object with just
        one sample representing the mean of all the samples in the original
        object.

    """
    return fdata.mean()


def var(fdatagrid):
    """Compute the variance of a set of samples in a FDataGrid object.

    Args:
        fdatagrid (FDataGrid): Object containing all the set of samples
        whose variance is desired.

    Returns:
        FDataGrid: A FDataGrid object with just one sample representing the
        variance of all the samples in the original FDataGrid object.

    """
    return fdatagrid.var()


def gmean(fdatagrid):
    """Compute the geometric mean of all the samples in a FDataGrid object.

    Args:
        fdatagrid (FDataGrid): Object containing all the samples whose
            geometric mean is wanted.

    Returns:
        FDataGrid: A FDataGrid object with just one sample representing the
        geometric mean of all the samples in the original FDataGrid object.

    """
    return fdatagrid.gmean()


def cov(fdatagrid):
    """Compute the covariance.

    Calculates the covariance matrix representing the covariance of the
    functional samples at the observation points.

    Args:
        fdatagrid (FDataGrid): Object containing different samples of a
            functional variable.

    Returns:
        numpy.darray: Matrix of covariances.

    """
    return fdatagrid.cov()



"""Discretised functional data module.

This module defines a class for representing functional data as a series of
lists of values, each representing the observation of a function measured in a
list of discretisation points.

"""

import numbers

import copy
import numpy
import scipy.stats.mstats


from . import basis as fdbasis
from .interpolation import SplineInterpolator
from . import FData
from .._utils import _list_of_arrays


__author__ = "Miguel Carbajo Berrocal"
__email__ = "miguel.carbajo@estudiante.uam.es"


class FDataGrid(FData):
    r"""Represent discretised functional data.

    Class for representing functional data as a set of curves discretised
    in a grid of points.

    Attributes:
        data_matrix (numpy.ndarray): a matrix where each entry of the first
            axis contains the values of a functional datum evaluated at the
            points of discretisation.
        sample_points (numpy.ndarray): 2 dimension matrix where each row
            contains the points of dicretisation for each axis of data_matrix.
        domain_range (numpy.ndarray): 2 dimension matrix where each row
            contains the bounds of the interval in which the functional data
            is considered to exist for each one of the axies.
        dataset_label (str): name of the dataset.
        axes_labels (list): list containing the labels of the different
            axis.
        extrapolation (str or Extrapolation): defines the default type of
            extrapolation. By default None, which does not apply any type of
            extrapolation. See `Extrapolation` for detailled information of the
            types of extrapolation.
        interpolator (GridInterpolator): Defines the type of interpolation
            applied in `evaluate`.
        keepdims (bool):

    Examples:
        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> data_matrix = [[1, 2], [2, 3]]
        >>> sample_points = [2, 4]
        >>> FDataGrid(data_matrix, sample_points)
        FDataGrid(
            array([[[1],
                    [2]],
        <BLANKLINE>
                   [[2],
                    [3]]]),
            sample_points=[array([2, 4])],
            domain_range=array([[2, 4]]),
            ...)

        The number of columns of data_matrix have to be the length of
        sample_points.

        >>> FDataGrid(numpy.array([1,2,4,5,8]), range(6))
        Traceback (most recent call last):
            ....
        ValueError: Incorrect dimension in data_matrix and sample_points...


        FDataGrid support higher dimensional data both in the domain and image.
        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}^2`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> sample_points = [2, 4]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> fd.ndim_domain, fd.ndim_image
        (1, 2)

        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}^2\longmapsto\mathbb{R}`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> sample_points = [[2, 4], [3,6]]
        >>> fd = FDataGrid(data_matrix, sample_points)
        >>> fd.ndim_domain, fd.ndim_image
        (2, 1)

    """

    def __init__(self, data_matrix, sample_points=None,
                 domain_range=None, dataset_label=None,
                 axes_labels=None, extrapolation=None,
                 interpolator=None, keepdims=False):

        """Construct a FDataGrid object.

        Args:
            data_matrix (array_like): a matrix where each row contains the
                values of a functional datum evaluated at the
                points of discretisation.
            sample_points (array_like, optional): an array containing the
                points of discretisation where values have been recorded or a list
                of lists with each of the list containing the points of
                dicretisation for each axis.
            domain_range (tuple or list of tuples, optional): contains the
                edges of the interval in which the functional data is
                considered to exist (if the argument has 2 dimensions each
                row is interpreted as the limits of one of the dimension of
                the domain).
            dataset_label (str, optional): name of the dataset.
            axes_labels (list, optional): list containing the labels of the
                different axes. The length of the list must be equal to the sum of the
                number of dimensions of the domain plus the number of dimensions
                of the image.

        """
        self.data_matrix = numpy.atleast_2d(data_matrix)

        if sample_points is None:
            self.sample_points = _list_of_arrays(
                [numpy.linspace(0, 1, self.data_matrix.shape[i]) for i in
                 range(1, self.data_matrix.ndim)])

        else:
            # Check that the dimension of the data matches the sample_points
            # list

            self.sample_points = _list_of_arrays(sample_points)

            data_shape = self.data_matrix.shape[1: 1 + self.ndim_domain]
            sample_points_shape = [len(i) for i in self.sample_points]

            if not numpy.array_equal(data_shape, sample_points_shape):
                raise ValueError("Incorrect dimension in data_matrix and "
                                 "sample_points. Data has shape {} and sample "
                                 "points have shape {}"
                                 .format(data_shape, sample_points_shape))


        self._sample_range = numpy.array(
            [(self.sample_points[i][0], self.sample_points[i][-1])
             for i in range(self.ndim_domain)])

        if domain_range is None:
                self._domain_range = self.sample_range
            # Default value for domain_range is a list of tuples with
            # the first and last element of each list ofthe sample_points.
        else:
            self._domain_range = numpy.atleast_2d(domain_range)
            # sample range must by a 2 dimension matrix with as many rows as
            # dimensions in the domain and 2 columns
            if (self._domain_range.ndim != 2 or self._domain_range.shape[1] != 2
                    or self._domain_range.shape[0] != self.ndim_domain):
                raise ValueError("Incorrect shape of domain_range.")
            for i in range(self.ndim_domain):
                if (self._domain_range[i, 0] > self.sample_points[i][0]
                        or self._domain_range[i, -1] < self.sample_points[i]
                        [-1]):
                    raise ValueError("Sample points must be within the domain "
                                     "range.")

        # Adjust the data matrix if the dimension of the image is one
        if self.data_matrix.ndim == 1 + self.ndim_domain:
            self.data_matrix = self.data_matrix[..., numpy.newaxis]

        if axes_labels is not None and len(axes_labels) != (self.ndim_domain + self.ndim_image):
            raise ValueError("There must be a label for each of the"
                              "dimensions of the domain and the image.")

        self.interpolator = interpolator


        super().__init__(extrapolation, dataset_label, axes_labels, keepdims)


        return

    def round(self, decimals=0):
        """Evenly round to the given number of decimals.

        Args:
            decimals (int, optional): Number of decimal places to round to.
                If decimals is negative, it specifies the number of
                positions to the left of the decimal point. Defaults to 0.

        Returns:
            :obj:FDataGrid: Returns a FDataGrid object where all elements
            in its data_matrix are rounded .The real and
            imaginary parts of complex numbers are rounded separately.

        """
        return self.copy(data_matrix=self.data_matrix.round(decimals))

    @property
    def ndim_domain(self):
        """Return number of dimensions of the domain.

        Returns:
            int: Number of dimensions of the domain.

        """
        return len(self.sample_points)

    @property
    def ndim_image(self):
        """Return number of dimensions of the image.

        Returns:
            int: Number of dimensions of the image.

        """
        try:
            # The dimension of the image is the length of the array that can
            #  be extracted from the data_matrix using all the dimensions of
            #  the domain.
            return self.data_matrix.shape[1 + self.ndim_domain]
        # If there is no array that means the dimension of the image is 1.
        except IndexError:
            return 1

    @property
    def ndim(self):
        """Return number of dimensions of the data matrix.

        Returns:
            int: Number of dimensions of the data matrix.

        """
        return self.data_matrix.ndim

    @property
    def nsamples(self):
        """Return number of rows of the data_matrix. Also the number of samples.

        Returns:
            int: Number of samples of the FDataGrid object. Also the number of
                rows of the data_matrix.

        """
        return self.data_matrix.shape[0]

    @property
    def ncol(self):
        """Return number of columns of the data_matrix.

        Also the number of points of discretisation.

        Returns:
            int: Number of columns of the data_matrix.

        """
        return self.data_matrix.shape[1]

    @property
    def sample_range(self):
        """Return the edges of the interval in which the functional data is
            considered to exist by the sample points.

            Do not have to be equal to the domain_range.
        """
        return self._sample_range

    @property
    def domain_range(self):
        """Return the edges of the interval in which the functional data is
            considered to exist by the sample points.

            Do not have to be equal to the sample_range.
        """
        return self._domain_range

    @property
    def shape(self):
        """Dimensions (aka shape) of the data_matrix.

        Returns:
            list of int: List containing the length of the matrix on each of
            its axis. If the matrix is 2 dimensional shape returns [number of
            rows, number of columns].

        """
        return self.data_matrix.shape

    @property
    def interpolator(self):
        """Defines the type of interpolation applied in `evaluate`."""
        return self._interpolator

    @interpolator.setter
    def interpolator(self, new_interpolator):
        """Sets the interpolator of the FDataGrid."""
        if new_interpolator is None:
            new_interpolator = SplineInterpolator()

        self._interpolator = new_interpolator
        self._interpolator_evaluator = None

    @property
    def _evaluator(self):
        """Return the evaluator constructed by the interpolator."""

        if self._interpolator_evaluator is None:
            self._interpolator_evaluator = self._interpolator.evaluator(self)

        return self._interpolator_evaluator


    def _evaluate(self, eval_points, *, derivative=0):
        """"Evaluate the object or its derivatives at a list of values.

        Args:
            eval_points (array_like): List of points where the functions are
                evaluated. If a matrix of shape nsample x eval_points is given
                each sample is evaluated at the values in the corresponding row
                in eval_points.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (numpy.darray): Matrix whose rows are the values of the each
            function at the values specified in eval_points.

        """

        return self._evaluator.evaluate(eval_points, derivative=derivative)

    def _evaluate_composed(self, eval_points, *, derivative=0):
        """"Evaluate the object or its derivatives at a list of values.

        Args:
            eval_points (array_like): List of points where the functions are
                evaluated. If a matrix of shape nsample x eval_points is given
                each sample is evaluated at the values in the corresponding row
                in eval_points.
            derivative (int, optional): Order of the derivative. Defaults to 0.

        Returns:
            (numpy.darray): Matrix whose rows are the values of the each
            function at the values specified in eval_points.

        """

        return self._evaluator.evaluate_composed(eval_points,
                                                 derivative=derivative)

    def derivative(self, order=1):
        r"""Differentiate a FDataGrid object.

        It is calculated using lagged differences. If we call :math:`D` the
        data_matrix, :math:`D^1` the derivative of order 1 and :math:`T` the
        vector contaning the points of discretisation; :math:`D^1` is
        calculated as it follows:

        .. math::

            D^{1}_{ij} = \begin{cases}
            \frac{D_{i1} - D_{i2}}{ T_{1} - T_{2}}  & \mbox{if } j = 1 \\
            \frac{D_{i(m-1)} - D_{im}}{ T_{m-1} - T_m}  & \mbox{if }
                j = m \\
            \frac{D_{i(j-1)} - D_{i(j+1)}}{ T_{j-1} - T_{j+1}} & \mbox{if }
            1 < j < m
            \end{cases}

        Where m is the number of columns of the matrix :math:`D`.

        Order > 1 derivatives are calculated by using derivative recursively.

        Args:
            order (int, optional): Order of the derivative. Defaults to one.

        Examples:
            First order derivative

            >>> fdata = FDataGrid([1,2,4,5,8], range(5))
            >>> fdata.derivative()
            FDataGrid(
                array([[[ 1. ],
                        [ 1.5],
                        [ 1.5],
                        [ 2. ],
                        [ 3. ]]]),
                sample_points=[array([0, 1, 2, 3, 4])],
                domain_range=array([[0, 4]]),
                ...)

            Second order derivative

            >>> fdata = FDataGrid([1,2,4,5,8], range(5))
            >>> fdata.derivative(2)
            FDataGrid(
                array([[[ 0.5 ],
                        [ 0.25],
                        [ 0.25],
                        [ 0.75],
                        [ 1.  ]]]),
                sample_points=[array([0, 1, 2, 3, 4])],
                domain_range=array([[0, 4]]),
                ...)

        """
        if self.ndim_domain != 1:
            raise NotImplementedError(
                "This method only works when the dimension "
                "of the domain of the FDatagrid object is "
                "one.")
        if order < 1:
            raise ValueError("The order of a derivative has to be greater "
                             "or equal than 1.")
        if self.ndim_domain > 1 or self.ndim_image > 1:
            raise NotImplementedError("Not implemented for 2 or more"
                                      " dimensional data.")
        if numpy.isnan(self.data_matrix).any():
            raise ValueError("The FDataGrid object cannot contain nan "
                             "elements.")
        data_matrix = self.data_matrix[..., 0]
        sample_points = self.sample_points[0]
        for _ in range(order):
            mdata = []
            for i in range(self.nsamples):
                arr = (numpy.diff(data_matrix[i]) /
                       (sample_points[1:]
                        - sample_points[:-1]))
                arr = numpy.append(arr, arr[-1])
                arr[1:-1] += arr[:-2]
                arr[1:-1] /= 2
                mdata.append(arr)
            data_matrix = numpy.array(mdata)

        if self.dataset_label:
            dataset_label = "{} - {} derivative".format(self.dataset_label,
                                                        order)
        else:
            dataset_label = None

        return self.copy(data_matrix=data_matrix, sample_points=sample_points,
                         dataset_label=dataset_label)

    def __check_same_dimensions(self, other):
        if self.data_matrix.shape[1] != other.data_matrix.shape[1]:
            raise ValueError("Error in columns dimensions")
        if not numpy.array_equal(self.sample_points, other.sample_points):
            raise ValueError("Sample points for both objects must be equal")

    def mean(self):
        """Compute the mean of all the samples.

        Returns:
            FDataGrid : A FDataGrid object with just one sample representing
            the mean of all the samples in the original object.

        """
        return self.copy(data_matrix=self.data_matrix.mean(axis=0,
                                                           keepdims=True))

    def var(self):
        """Compute the variance of a set of samples in a FDataGrid object.

        Returns:
            FDataGrid: A FDataGrid object with just one sample representing the
            variance of all the samples in the original FDataGrid object.

        """
        return self.copy(data_matrix=[numpy.var(self.data_matrix, 0)])

    def cov(self):
        """Compute the covariance.

        Calculates the covariance matrix representing the covariance of the
        functional samples at the observation points.

        Returns:
            numpy.darray: Matrix of covariances.

        """

        if self.dataset_label is not None:
            dataset_label = self.dataset_label + ' - covariance'
        else:
            dataset_label = None

        return self.copy(data_matrix=numpy.cov(self.data_matrix,
                                               rowvar=False)[numpy.newaxis, ...],
                         sample_points=[self.sample_points[0],
                                        self.sample_points[0]],
                         domain_range=[self.domain_range[0],
                                       self.domain_range[0]],
                         dataset_label=dataset_label)

    def gmean(self):
        """Compute the geometric mean of all samples in the FDataGrid object.

        Returns:
            FDataGrid: A FDataGrid object with just one sample representing
            the geometric mean of all the samples in the original
            FDataGrid object.

        """
        return self.copy(data_matrix=
                         [scipy.stats.mstats.gmean(self.data_matrix, 0)])

    def __add__(self, other):
        """Addition for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        if isinstance(other, (numpy.ndarray, numbers.Number)):
            data_matrix = other
        elif isinstance(other, FDataGrid):
            self.__check_same_dimensions(other)
            data_matrix = other.data_matrix
        else:
            return NotImplemented

        return self.copy(data_matrix=self.data_matrix + data_matrix)

    def __radd__(self, other):
        """Addition for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """

        return self.__add__(other)

    def __sub__(self, other):
        """Subtraction for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        if isinstance(other, (numpy.ndarray, numbers.Number)):
            data_matrix = other
        elif isinstance(other, FDataGrid):
            self.__check_same_dimensions(other)
            data_matrix = other.data_matrix
        else:
            return NotImplemented

        return self.copy(data_matrix=self.data_matrix - data_matrix)

    def __rsub__(self, other):
        """Right Subtraction for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        if isinstance(other, (numpy.ndarray, numbers.Number)):
            data_matrix = other
        elif isinstance(other, FDataGrid):
            self.__check_same_dimensions(other)
            data_matrix = other.data_matrix
        else:
            return NotImplemented

        return self.copy(data_matrix=data_matrix - self.data_matrix)

    def __mul__(self, other):
        """Multiplication for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        if isinstance(other, (numpy.ndarray, numbers.Number)):
            data_matrix = other
        elif isinstance(other, FDataGrid):
            self.__check_same_dimensions(other)
            data_matrix = other.data_matrix
        else:
            return NotImplemented

        return self.copy(data_matrix=self.data_matrix * data_matrix)

    def __rmul__(self, other):
        """Multiplication for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        return self.__mul__(other)

    def __truediv__(self, other):
        """Division for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        if isinstance(other, (numpy.ndarray, numbers.Number)):
            data_matrix = other
        elif isinstance(other, FDataGrid):
            self.__check_same_dimensions(other)
            data_matrix = other.data_matrix
        else:
            return NotImplemented

        return self.copy(data_matrix=self.data_matrix / data_matrix)

    def __rtruediv__(self, other):
        """Division for FDataGrid object.

        It supports other FDataGrid objects, numpy.ndarray and numbers.

        """
        if isinstance(other, (numpy.ndarray, numbers.Number)):
            data_matrix = other
        elif isinstance(other, FDataGrid):
            self.__check_same_dimensions(other)
            data_matrix = other.data_matrix
        else:
            return NotImplemented

        return self.copy(data_matrix=data_matrix / self.data_matrix)


    def concatenate(self, other):
        """Join samples from a similar FDataGrid object.

        Joins samples from another FDataGrid object if it has the same
        dimensions and sampling points.

        Args:
            other (:obj:`FDataGrid`): another FDataGrid object.

        Returns:
            :obj:`FDataGrid`: FDataGrid object with the samples from the two
            original objects.

        Examples:
            >>> fd = FDataGrid([1,2,4,5,8], range(5))
            >>> fd_2 = FDataGrid([3,4,7,9,2], range(5))
            >>> fd.concatenate(fd_2)
            FDataGrid(
                array([[[1],
                        [2],
                        [4],
                        [5],
                        [8]],
            <BLANKLINE>
                       [[3],
                        [4],
                        [7],
                        [9],
                        [2]]]),
                sample_points=[array([0, 1, 2, 3, 4])],
                domain_range=array([[0, 4]]),
                ...)

        """
        # Checks
        self.__check_same_dimensions(other)

        return self.copy(data_matrix=numpy.concatenate((self.data_matrix,
                                                        other.data_matrix),
                                                       axis=0))


    def scatter(self, fig=None, ax=None, nrows=None, ncols=None, **kwargs):
        """Scatter plot of the FDatGrid object.

        Args:
            fig (figure object, optional): figure over with the graphs are plotted in case ax is not specified.
                If None and ax is also None, the figure is initialized.
            ax (list of axis objects, optional): axis over where the graphs are plotted. If None, see param fig.
            nrows(int, optional): designates the number of rows of the figure to plot the different dimensions of the
                image. Only specified if fig and ax are None.
            ncols(int, optional): designates the number of columns of the figure to plot the different dimensions of the
                image. Only specified if fig and ax are None.
            **kwargs: keyword arguments to be passed to the matplotlib.pyplot.scatter function;

        Returns:
            fig (figure object): figure object in which the graphs are plotted in case ax is None.
            ax (axes object): axes in which the graphs are plotted.

        """
        fig, ax = self.generic_plotting_checks(fig, ax, nrows, ncols)

        if self.ndim_domain == 1:
            for i in range(self.ndim_image):
                for j in range(self.nsamples):
                    ax[i].scatter(self.sample_points[0], self.data_matrix[j, :, i].T, **kwargs)
        else:
            X = self.sample_points[0]
            Y = self.sample_points[1]
            X, Y = numpy.meshgrid(X, Y)
            for i in range(self.ndim_image):
                for j in range(self.nsamples):
                    ax[i].scatter(X, Y, self.data_matrix[j, :, :, i].T, **kwargs)

        self.set_labels(fig, ax)

        return fig, ax


    def to_basis(self, basis, **kwargs):
        """Return the basis representation of the object.

        Args:
            basis(Basis): basis object in which the functional data are
                going to be represented.
            **kwargs: keyword arguments to be passed to
                FDataBasis.from_data().

        Returns:
            FDataBasis: Basis representation of the funtional data
            object.

        Examples:
            >>> import numpy as np
            >>> import skfda
            >>> t = np.linspace(0, 1, 5)
            >>> x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
            >>> x
            array([ 1.,  1., -1., -1.,  1.])

            >>> fd = FDataGrid(x, t)
            >>> basis = skfda.representation.basis.Fourier((0, 1), nbasis=3)
            >>> fd_b = fd.to_basis(basis)
            >>> fd_b.coefficients.round(2)
            array([[ 0.  , 0.71, 0.71]])

        """
        if self.ndim_domain > 1:
            raise NotImplementedError("Only support 1 dimension on the "
                                      "domain.")
        elif self.ndim_image > 1:
            raise NotImplementedError("Only support 1 dimension on the "
                                      "image.")
        return fdbasis.FDataBasis.from_data(self.data_matrix[..., 0],
                                            self.sample_points[0],
                                            basis,
                                            keepdims=self.keepdims,
                                            **kwargs)

    def to_grid(self, sample_points=None):
        """Return the discrete representation of the object.

        Args:
            sample_points (array_like, optional):  2 dimension matrix where each
            row contains the points of dicretisation for each axis of
            data_matrix.

        Returns:
              FDataGrid: Discrete representation of the functional data
              object.

        """
        if sample_points is None:
            sample_points = self.sample_points

        return self.copy(data_matrix=self.evaluate(sample_points, grid=True),
                         sample_points=sample_points)



    def copy(self, *, data_matrix=None, sample_points=None,
                 domain_range=None, dataset_label=None,
                 axes_labels=None, extrapolation=None,
                 interpolator=None, keepdims=None):
        """Returns a copy of the FDataGrid.

        If an argument is provided the corresponding attribute in the new copy
        is updated.

        """

        if data_matrix is None:
            # The data matrix won't be writeable
            data_matrix = self.data_matrix

        if sample_points is None:
            # Sample points won`t be writeable
            sample_points = self.sample_points

        if domain_range is None:
            domain_range = copy.deepcopy(self.domain_range)

        if dataset_label is None:
            dataset_label = copy.copy(self.dataset_label)

        if axes_labels is None:
            axes_labels = copy.copy(self.axes_labels)

        if extrapolation is None:
            extrapolation = self.extrapolation

        if interpolator is None:
            interpolator = self.interpolator

        if keepdims is None:
            keepdims = self.keepdims

        return FDataGrid(data_matrix, sample_points=sample_points,
                     domain_range=domain_range, dataset_label=dataset_label,
                     axes_labels=axes_labels, extrapolation=extrapolation,
                     interpolator=interpolator, keepdims=keepdims)


    def shift(self, shifts, *, restrict_domain=False, extrapolation=None,
              eval_points=None):
        """Perform a shift of the curves.

        Args:
            shifts (array_like or numeric): List with the shifts
                corresponding for each sample or numeric with the shift to apply
                to all samples.
            restrict_domain (bool, optional): If True restricts the domain to
                avoid evaluate points outside the domain using extrapolation.
                Defaults uses extrapolation.
            extrapolation (str or Extrapolation, optional): Controls the
                extrapolation mode for elements outside the domain range.
                By default uses the method defined in fd. See extrapolation to
                more information.
            eval_points (array_like, optional): Set of points where
                the functions are evaluated to obtain the discrete
                representation of the object to operate. If an empty list the
                current sample_points are used to unificate the domain of the
                shifted data.

        Returns:
            :class:`FDataGrid` with the shifted data.
        """


        if numpy.isscalar(shifts):
            shifts = [shifts]

        shifts = numpy.array(shifts)

        # Case unidimensional treated as the multidimensional
        if self.ndim_domain == 1 and shifts.ndim == 1 and shifts.shape[0] != 1:
            shifts = shifts[:, numpy.newaxis]

        # Case same shift for all the curves
        if shifts.shape[0] == self.ndim_domain and shifts.ndim ==1:

            # Column vector with shapes
            shifts = numpy.atleast_2d(shifts).T

            sample_points = self.sample_points + shifts
            domain_range = self.domain_range + shifts

            return self.copy(sample_points=sample_points,
                             domain_range=domain_range)


        if shifts.shape[0] != self.nsamples:
            raise ValueError(f"shifts vector ({shifts.shape[0]}) must have the "
                             f"same length than the number of samples "
                             f"({self.nsamples})")

        if eval_points is None:
            eval_points = self.sample_points



        if restrict_domain:
            domain = numpy.asarray(self.domain_range)
            a = domain[:,0] - numpy.atleast_1d(numpy.min(numpy.min(shifts, axis=1), 0))
            b = domain[:,1] - numpy.atleast_1d(numpy.max(numpy.max(shifts, axis=1), 0))

            domain = numpy.vstack((a,b)).T

            eval_points = [eval_points[i][
                numpy.logical_and(eval_points[i] >= domain[i,0],
                                  eval_points[i] <= domain[i,1])]
                           for i in range(self.ndim_domain)]

        else:
            domain = self.domain_range

        eval_points = numpy.asarray(eval_points)


        eval_points_repeat = numpy.repeat(eval_points[numpy.newaxis, :],
                                       self.nsamples, axis=0)

        # Solve problem with cartesian and matrix indexing
        if self.ndim_domain > 1:
            shifts[:,:2] = numpy.flip(shifts[:,:2], axis=1)

        shifts = numpy.repeat(shifts[..., numpy.newaxis],
                              eval_points.shape[1], axis=2)

        eval_points_shifted = eval_points_repeat + shifts


        grid = True if self.ndim_domain > 1 else False

        data_matrix = self.evaluate(eval_points_shifted,
                                    extrapolation=extrapolation,
                                    aligned_evaluation=False,
                                    grid=True)


        return self.copy(data_matrix=data_matrix, sample_points=eval_points,
                         domain_range=domain)

    def compose(self, fd, *, eval_points=None):
        """Composition of functions.

        Performs the composition of functions.

        Args:
            fd (:class:`FData`): FData object to make the composition. Should
                have the same number of samples and image dimension equal to 1.
            eval_points (array_like): Points to perform the evaluation.
        """

        if self.ndim_domain != fd.ndim_image:
            raise ValueError(f"Dimension of codomain of first function do not "
                             f"match with the domain of the second function "
                             f"({self.ndim_domain})!=({fd.ndim_image}).")

        # All composed with same function
        if fd.nsamples == 1 and self.nsamples != 1:
            fd = fd.copy(data_matrix=numpy.repeat(fd.data_matrix, self.nsamples,
                                               axis=0))

        if fd.ndim_domain == 1:
            if eval_points is None:
                try:
                    eval_points = fd.sample_points[0]
                except:
                    eval_points = numpy.linspace(*fd.domain_range[0], 201)

            eval_points_transformation = fd(eval_points, keepdims=False)
            data_matrix = self(eval_points_transformation,
                               aligned_evaluation=False)
        else:
            if eval_points is None:
                eval_points = fd.sample_points

            grid_transformation = fd(eval_points, grid=True, keepdims=True)

            lengths = [len(ax) for ax in eval_points]

            eval_points_transformation =  numpy.empty((self.nsamples,
                                                       numpy.prod(lengths),
                                                       self.ndim_domain))


            for i in range(self.nsamples):
                eval_points_transformation[i] = numpy.array(
                    list(map(numpy.ravel, grid_transformation[i].T))
                    ).T

            data_flatten = self(eval_points_transformation,
                               aligned_evaluation=False)

            data_matrix = data_flatten.reshape((self.nsamples, *lengths,
                                                self.ndim_image))


        return self.copy(data_matrix=data_matrix,
                         sample_points=eval_points,
                         domain_range=fd.domain_range)



    def __str__(self):
        """Return str(self)."""
        return ('Data set:    ' + str(self.data_matrix)
                + '\nsample_points:    ' + str(self.sample_points)
                + '\ntime range:    ' + str(self.domain_range))

    def __repr__(self):
        """Return repr(self)."""
        return (f"FDataGrid("
                f"\n{repr(self.data_matrix)},"
                f"\nsample_points={repr(self.sample_points)},"
                f"\ndomain_range={repr(self.domain_range)},"
                f"\ndataset_label={repr(self.dataset_label)},"
                f"\naxes_labels={repr(self.axes_labels)},"
                f"\nextrapolation={repr(self.extrapolation)},"
                f"\ninterpolator={repr(self.interpolator)},"
                f"\nkeepdims={repr(self.keepdims)})").replace('\n', '\n    ')

    def __getitem__(self, key):
        """Return self[key]."""
        if isinstance(key, tuple):
            # If there are not values for every dimension, the remaining ones
            # are kept
            key += (slice(None),) * (self.ndim_domain + 1 - len(key))

            sample_points = [self.sample_points[i][subkey]
                             for i, subkey in enumerate(
                                 key[1:1 + self.ndim_domain])]

            return self.copy(data_matrix=self.data_matrix[key],
                             sample_points=sample_points)

        if isinstance(key, int):
            return self.copy(data_matrix=self.data_matrix[key:key + 1])

        else:
            return self.copy(data_matrix=self.data_matrix[key])

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        for i in inputs:
            if isinstance(i, FDataGrid) and not numpy.all(i.sample_points ==
                                                          self.sample_points):
                return NotImplemented

        new_inputs = [i.data_matrix if isinstance(i, FDataGrid)
                      else i for i in inputs]

        outputs = kwargs.pop('out', None)
        if outputs:
            new_outputs = [o.data_matrix if isinstance(o, FDataGrid)
                           else o for o in outputs]
            kwargs['out'] = tuple(new_outputs)
        else:
            new_outputs = (None,) * ufunc.nout

        results = getattr(ufunc, method)(*new_inputs, **kwargs)
        if results is NotImplemented:
            return NotImplemented

        if ufunc.nout == 1:
            results = (results,)

        results = tuple((result
                         if output is None else output)
                        for result, output in zip(results, new_outputs))

        results = [self.copy(data_matrix=r) for r in results]

        return results[0] if len(results) == 1 else results



import unittest

from skfda.representation.basis import (Basis, FDataBasis, Constant, Monomial,
                                        BSpline, Fourier)

import numpy as np


class TestBasis(unittest.TestCase):

    # def setUp(self): could be defined for set up before any test

    def test_from_data_cholesky(self):
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSpline((0, 1), nbasis=5)
        np.testing.assert_array_almost_equal(
            FDataBasis.from_data(x, t, basis, smoothness_parameter=10,
                                 penalty_degree=2, method='cholesky'
                                 ).coefficients.round(2),
            np.array([[0.60, 0.47, 0.20, -0.07, -0.20]])
        )

    def test_from_data_qr(self):
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSpline((0, 1), nbasis=5)
        np.testing.assert_array_almost_equal(
            FDataBasis.from_data(x, t, basis, smoothness_parameter=10,
                                 penalty_degree=2, method='qr'
                                 ).coefficients.round(2),
            np.array([[0.60, 0.47, 0.20, -0.07, -0.20]])
        )

    def test_monomial_smoothing(self):
        # It does not have much sense to apply smoothing in this basic case
        # where the fit is very good but its just for testing purposes
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = Monomial(nbasis=4)
        fd = FDataBasis.from_data(x, t, basis,
                                  penalty_degree=2,
                                  smoothness_parameter=1)
        # These results where extracted from the R package fda
        np.testing.assert_array_almost_equal(
            fd.coefficients.round(2), np.array([[0.61, -0.88, 0.06, 0.02]]))

    def test_bspline_penalty_special_case(self):
        basis = BSpline(nbasis=5)
        np.testing.assert_array_almost_equal(
            basis.penalty(basis.order - 1),
            np.array([[1152., -2016., 1152., -288., 0.],
                      [-2016., 3600., -2304., 1008., -288.],
                      [1152., -2304., 2304., -2304., 1152.],
                      [-288., 1008., -2304., 3600., -2016.],
                      [0., -288., 1152., -2016., 1152.]]))

    def test_fourier_penalty(self):
        basis = Fourier(nbasis=5)
        np.testing.assert_array_almost_equal(
            basis.penalty(2).round(2),
            np.array([[0., 0., 0., 0., 0.],
                      [0., 1558.55, 0., 0., 0.],
                      [0., 0., 1558.55, 0., 0.],
                      [0., 0., 0., 24936.73, 0.],
                      [0., 0., 0., 0., 24936.73]]))

    def test_bspline_penalty(self):
        basis = BSpline(nbasis=5)
        np.testing.assert_array_almost_equal(
            basis.penalty(2).round(2),
            np.array([[96., -132., 24., 12., 0.],
                      [-132., 192., -48., -24., 12.],
                      [24., -48., 48., -48., 24.],
                      [12., -24., -48., 192., -132.],
                      [0., 12., 24., -132., 96.]]))

    def test_bspline_penalty_numerical(self):
        basis = BSpline(nbasis=5)
        np.testing.assert_array_almost_equal(
            basis.penalty(coefficients=[0, 0, 1]).round(2),
            np.array([[96., -132., 24., 12., 0.],
                      [-132., 192., -48., -24., 12.],
                      [24., -48., 48., -48., 24.],
                      [12., -24., -48., 192., -132.],
                      [0., 12., 24., -132., 96.]]))

    def test_basis_product_generic(self):
        monomial = Monomial(nbasis=5)
        fourier = Fourier(nbasis=3)
        prod = BSpline(nbasis=9, order=8)
        self.assertEqual(Basis.default_basis_of_product(monomial, fourier), prod)

    def test_basis_constant_product(self):
        constant = Constant()
        monomial = Monomial()
        fourier = Fourier()
        bspline = BSpline(nbasis=5, order=3)
        self.assertEqual(constant.basis_of_product(monomial), monomial)
        self.assertEqual(constant.basis_of_product(fourier), fourier)
        self.assertEqual(constant.basis_of_product(bspline), bspline)
        self.assertEqual(monomial.basis_of_product(constant), monomial)
        self.assertEqual(fourier.basis_of_product(constant), fourier)
        self.assertEqual(bspline.basis_of_product(constant), bspline)

    def test_basis_fourier_product(self):
        # Test when periods are the same
        fourier = Fourier(nbasis=5)
        fourier2 = Fourier(nbasis=3)
        prod = Fourier(nbasis=7)
        self.assertEqual(fourier.basis_of_product(fourier2), prod)

        # Test when periods are different
        fourier2 = Fourier(nbasis=3, period=2)
        prod = BSpline(nbasis=9, order=8)
        self.assertEqual(fourier.basis_of_product(fourier2), prod)

    def test_basis_monomial_product(self):
        monomial = Monomial(nbasis=5)
        monomial2 = Monomial(nbasis=3)
        prod = Monomial(nbasis=8)
        self.assertEqual(monomial.basis_of_product(monomial2), prod)

    def test_basis_bspline_product(self):
        bspline = BSpline(nbasis=6, order=4)
        bspline2 = BSpline(domain_range=(0, 1), nbasis=6, order=4, knots=[0, 0.3, 1 / 3, 1])
        prod = BSpline(domain_range=(0,1), nbasis=10, order=7, knots=[0, 0.3, 1/3, 2/3,1])
        self.assertEqual(bspline.basis_of_product(bspline2), prod)

    def test_basis_basis_inprod(self):
        monomial = Monomial(nbasis=4)
        bspline = BSpline(nbasis=5, order=4)
        np.testing.assert_array_almost_equal(
            monomial.inner_product(bspline).round(3),
            np.array(
                [[0.12499983, 0.25000035, 0.24999965, 0.25000035, 0.12499983],
                 [0.01249991, 0.07500017, 0.12499983, 0.17500017, 0.11249991],
                 [0.00208338, 0.02916658, 0.07083342, 0.12916658, 0.10208338],
                 [0.00044654, 0.01339264, 0.04375022, 0.09910693, 0.09330368]])
            .round(3)
        )

    def test_basis_fdatabasis_inprod(self):
        monomial = Monomial(nbasis=4)
        bspline = BSpline(nbasis=5, order=3)
        bsplinefd = FDataBasis(bspline, np.arange(0, 15).reshape(3, 5))

        np.testing.assert_array_almost_equal(
            monomial.inner_product(bsplinefd).round(3),
            np.array([[2., 7., 12.],
                      [1.29626206, 3.79626206, 6.29626206],
                      [0.96292873, 2.62959539, 4.29626206],
                      [0.7682873, 2.0182873, 3.2682873]]).round(3)
        )

    def test_fdatabasis_fdatabasis_inprod(self):
        monomial = Monomial(nbasis=4)
        monomialfd = FDataBasis(monomial, [[5, 4, 1, 0],
                                           [4, 2, 1, 0],
                                           [4, 1, 6, 4],
                                           [4, 5, 0, 1],
                                           [5, 6, 2, 0]])
        bspline = BSpline(nbasis=5, order=3)
        bsplinefd = FDataBasis(bspline, np.arange(0, 15).reshape(3, 5))

        np.testing.assert_array_almost_equal(
            monomialfd.inner_product(bsplinefd).round(3),
            np.array([[16.14797697, 52.81464364, 89.4813103],
                      [11.55565285, 38.22211951, 64.88878618],
                      [18.14698361, 55.64698361, 93.14698361],
                      [15.2495976, 48.9995976, 82.7495976],
                      [19.70392982, 63.03676315, 106.37009648]]).round(3)
        )

    def test_comutativity_inprod(self):
        monomial = Monomial(nbasis=4)
        bspline = BSpline(nbasis=5, order=3)
        bsplinefd = FDataBasis(bspline, np.arange(0, 15).reshape(3, 5))

        np.testing.assert_array_almost_equal(
            bsplinefd.inner_product(monomial).round(3),
            np.transpose(monomial.inner_product(bsplinefd).round(3))
        )

    def test_fdatabasis_times_fdatabasis_fdatabasis(self):
        monomial = FDataBasis(Monomial(nbasis=3), [1, 2, 3])
        bspline = FDataBasis(BSpline(nbasis=6, order=4), [1, 2, 4, 1, 0, 1])
        times_fdar = monomial.times(bspline)

        prod_basis = BSpline(nbasis=9, order=6, knots=[0, 0.25, 0.5, 0.75, 1])
        prod_coefs = np.array([[0.9788352,  1.6289955,  2.7004969,  6.2678739,
                      8.7636441,  4.0069960,  0.7126961,  2.8826708,
                      6.0052311]])

        self.assertEqual(prod_basis, times_fdar.basis)
        np.testing.assert_array_almost_equal(prod_coefs, times_fdar.coefficients)

    def test_fdatabasis_times_fdatabasis_list(self):
        monomial = FDataBasis(Monomial(nbasis=3),
                              [[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        result = monomial.times([3, 2, 1])

        expec_basis = Monomial(nbasis=3)
        expec_coefs = np.array([[3, 6, 9], [8, 10, 12], [7, 8, 9]])

        self.assertEqual(expec_basis, result.basis)
        np.testing.assert_array_almost_equal(expec_coefs, result.coefficients)

    def test_fdatabasis_times_fdatabasis_int(self):
        monomial = FDataBasis(Monomial(nbasis=3),
                              [[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        result = monomial.times(3)

        expec_basis = Monomial(nbasis=3)
        expec_coefs = np.array([[3, 6, 9], [12, 15, 18], [21, 24, 27]])

        self.assertEqual(expec_basis, result.basis)
        np.testing.assert_array_almost_equal(expec_coefs, result.coefficients)

    def test_fdatabasis_derivative_constant(self):
        monomial = FDataBasis(Monomial(nbasis=8),
                              [1, 5, 8, 9, 7, 8, 4, 5])
        monomial2 = FDataBasis(Monomial(nbasis=5),
                              [[4, 9, 7, 4, 3],
                               [1, 7, 9, 8, 5],
                               [4, 6, 6, 6, 8]])

        np.testing.assert_equal(monomial.derivative(),
                                FDataBasis(Monomial(nbasis=7),
                                           [5, 16, 27, 28, 40, 24, 35]))
        np.testing.assert_equal(monomial.derivative(order=0), monomial)
        np.testing.assert_equal(monomial.derivative(order=6),
                                FDataBasis(Monomial(nbasis=2),
                                           [2880, 25200]))

        np.testing.assert_equal(monomial2.derivative(),
                                FDataBasis(Monomial(nbasis=4),
                                           [[9, 14, 12, 12],
                                            [7, 18, 24, 20],
                                            [6, 12, 18, 32]]))
        np.testing.assert_equal(monomial2.derivative(order=0), monomial2)
        np.testing.assert_equal(monomial2.derivative(order=3),
                                FDataBasis(Monomial(nbasis=2),
                                           [[24, 72],
                                            [48, 120],
                                            [36, 192]]))

    def test_fdatabasis_derivative_monomial(self):
        monomial = FDataBasis(Monomial(nbasis=8),
                              [1, 5, 8, 9, 7, 8, 4, 5])
        monomial2 = FDataBasis(Monomial(nbasis=5),
                              [[4, 9, 7, 4, 3],
                               [1, 7, 9, 8, 5],
                               [4, 6, 6, 6, 8]])

        np.testing.assert_equal(monomial.derivative(),
                                FDataBasis(Monomial(nbasis=7),
                                           [5, 16, 27, 28, 40, 24, 35]))
        np.testing.assert_equal(monomial.derivative(order=0), monomial)
        np.testing.assert_equal(monomial.derivative(order=6),
                                FDataBasis(Monomial(nbasis=2),
                                           [2880, 25200]))

        np.testing.assert_equal(monomial2.derivative(),
                                FDataBasis(Monomial(nbasis=4),
                                           [[9, 14, 12, 12],
                                            [7, 18, 24, 20],
                                            [6, 12, 18, 32]]))
        np.testing.assert_equal(monomial2.derivative(order=0), monomial2)
        np.testing.assert_equal(monomial2.derivative(order=3),
                                FDataBasis(Monomial(nbasis=2),
                                           [[24, 72],
                                            [48, 120],
                                            [36, 192]]))

    def test_fdatabasis_derivative_fourier(self):
        fourier = FDataBasis(Fourier(nbasis=7),
                              [1, 5, 8, 9, 8, 4, 5])
        fourier2 = FDataBasis(Fourier(nbasis=5),
                              [[4, 9, 7, 4, 3],
                               [1, 7, 9, 8, 5],
                               [4, 6, 6, 6, 8]])

        fou0 = fourier.derivative(order=0)
        fou1 = fourier.derivative()
        fou2 = fourier.derivative(order=2)

        np.testing.assert_equal(fou1.basis, fourier.basis)
        np.testing.assert_almost_equal(fou1.coefficients.round(5),
                                       np.atleast_2d([0, -50.26548, 31.41593,
                                                      -100.53096, 113.09734,
                                                      -94.24778, 75.39822]))
        np.testing.assert_equal(fou0, fourier)
        np.testing.assert_equal(fou2.basis, fourier.basis)
        np.testing.assert_almost_equal(fou2.coefficients.round(5),
                                       np.atleast_2d([0, -197.39209, -315.82734,
                                                      -1421.22303, -1263.30936,
                                                      -1421.22303, -1776.52879]))

        fou0 = fourier2.derivative(order=0)
        fou1 = fourier2.derivative()
        fou2 = fourier2.derivative(order=2)

        np.testing.assert_equal(fou1.basis, fourier2.basis)
        np.testing.assert_almost_equal(fou1.coefficients.round(5),
                                       [[0, -43.98230, 56.54867, -37.69911, 50.26548],
                                        [0, -56.54867, 43.98230, -62.83185, 100.53096],
                                        [0, -37.69911, 37.69911, -100.53096, 75.39822]])
        np.testing.assert_equal(fou0, fourier2)
        np.testing.assert_equal(fou2.basis, fourier2.basis)
        np.testing.assert_almost_equal(fou2.coefficients.round(5),
                                       [[0, -355.30576, -276.34892, -631.65468, -473.74101],
                                        [0, -276.34892, -355.30576, -1263.30936, -789.56835],
                                        [0, -236.87051, -236.87051, -947.48202, -1263.30936]])

    def test_fdatabasis_derivative_bspline(self):
        bspline = FDataBasis(BSpline(nbasis=8),
                             [1, 5, 8, 9, 7, 8, 4, 5])
        bspline2 = FDataBasis(BSpline(nbasis=5),
                              [[4, 9, 7, 4, 3],
                               [1, 7, 9, 8, 5],
                               [4, 6, 6, 6, 8]])

        bs0 = bspline.derivative(order=0)
        bs1 = bspline.derivative()
        bs2 = bspline.derivative(order=2)
        np.testing.assert_equal(bs1.basis, BSpline(nbasis=7, order=3))
        np.testing.assert_almost_equal(bs1.coefficients,
                                       np.atleast_2d([60, 22.5, 5,
                                                      -10, 5, -30, 15]))
        np.testing.assert_equal(bs0, bspline)
        np.testing.assert_equal(bs2.basis, BSpline(nbasis=6, order=2))
        np.testing.assert_almost_equal(bs2.coefficients,
                                       np.atleast_2d([-375, -87.5, -75,
                                                      75, -175, 450]))

        bs0 = bspline2.derivative(order=0)
        bs1 = bspline2.derivative()
        bs2 = bspline2.derivative(order=2)

        np.testing.assert_equal(bs1.basis, BSpline(nbasis=4, order=3))
        np.testing.assert_almost_equal(bs1.coefficients,
                                       [[30, -6, -9, -6],
                                        [36, 6, -3, -18],
                                        [12, 0, 0, 12]])
        np.testing.assert_equal(bs0, bspline2)
        np.testing.assert_equal(bs2.basis, BSpline(nbasis=3, order=2))
        np.testing.assert_almost_equal(bs2.coefficients,
                                       [[-144, -6, 12],
                                        [-120, -18, -60],
                                        [-48, 0, 48]])


if __name__ == '__main__':
    print()
    unittest.main()
